import pandas as pd
import numpy as np
import seaborn as sns
import matplotlib.pyplot as plt

import tensorflow as tf
import joblib, sys, os
from sklearn.metrics import f1_score, confusion_matrix, roc_auc_score, roc_curve, classification_report
from tensorflow.keras.models import Sequential, Model
from tensorflow.keras.layers import Conv1D, MaxPooling1D, Flatten, Dense, Dropout, Input, LeakyReLU, concatenate, BatchNormalization
from tensorflow.keras.optimizers import Adam
import mlflow
from mlflow.models import ModelSignature
from mlflow.types import Schema, TensorSpec
from keras.callbacks import ReduceLROnPlateau
from datetime import datetime
from tensorflow.python.client import device_lib


print("device LIST    ", device_lib.list_local_devices())
gpus = tf.config.experimental.list_physical_devices('GPU')
print("Num GPUs Available: ", len(gpus))
for gpu in gpus:
  tf.config.experimental.set_memory_growth(gpu, True)
print("TensorFlow version:", tf.__version__)

def convert_to_string(lst):
    if not lst:
        return ''
    return ', '.join(lst)

def switch_places(entry):
    if ',' in entry:
        parts = entry.split(',')
        parts = [word.strip() for word in parts]
        parts.sort()  
        return ', '.join(parts)
    return entry

def load_data(directory, file_name):
    path = os.path.join(directory, file_name)
    return joblib.load(path)

def prepare_data_classes(yt, yts, yv, class1, class0):
        
    for item in yt:
        item['label'] = 1 if class1 in item['label'] else 0
    for item in yts:
        item['label'] = 1 if class1 in item['label'] else 0
    for item in yv:
        item['label'] = 1 if class1 in item['label'] else 0

    yt = [item for item in yt if item['label'] != 2]
    yts = [item for item in yts if item['label'] != 2]
    yv = [item for item in yv if item['label'] != 2]

    y_tr = np.array([item['label'] for item in yt])
    y_te = np.array([item['label'] for item in yts])
    y_va = np.array([item['label'] for item in yv])

    return y_tr, y_te, y_va

if __name__ == "__main__":
    
    input_directory1 = "../records_ptbxl"  
    output_directory = "reruns/HYP_binary_classification"

    if not os.path.exists(output_directory):
        os.makedirs(output_directory)
    leads = [] # odje navodimo kanale (indekse od 0 do 11) koje zelimo da koristimo; ako je lista prazna podrazumijeva se da svaki kanala ucestvuje
    class0 = "OTHERS"
    class1 = "HYP"

    X1 = load_data(input_directory1, "X_train.sav") 
    X1 = np.array([item['signal'] for item in X1])
    y_train = load_data(input_directory1, "y_train.sav") 
    features_train = load_data(input_directory1, "features_train.sav")
    features_train = np.array([item['features'] for item in features_train])

    X2 = load_data(input_directory1, "X_test.sav") 
    X2 = np.array([item['signal'] for item in X2])
    y_test = load_data(input_directory1, "y_test.sav") 
    features_test = load_data(input_directory1, "features_test.sav")
    features_test = np.array([item['features'] for item in features_test])

    X3 = load_data(input_directory1, "X_val.sav") 
    X3 = np.array([item['signal'] for item in X3])
    y_val = load_data(input_directory1, "y_val.sav") 
    features_val = load_data(input_directory1, "features_val.sav")
    features_val = np.array([item['features'] for item in features_val])

    y_train, y_test, y_val = prepare_data_classes(y_train, y_test, y_val, class1, class0)


    with tf.device('/GPU:0'):
        features_train = np.asarray(features_train).astype('float32')
        features_test = np.asarray(features_test).astype('float32')
        features_val = np.asarray(features_val).astype('float32')
    
        num_samp = 1000 # **************
        ls = len(leads)
        if(ls == 0):
            num_leads = 12
        else:
            num_leads = ls
        X_train = np.zeros((X1.shape[0], num_samp, num_leads))
        X_test = np.zeros((X2.shape[0], num_samp, num_leads))
        X_val = np.zeros((X3.shape[0], num_samp, num_leads))
    
    ###########################################################################
    # ovdje se selektuju kanali koji ce ucestvovati u treniranju i testiranju
        if ls != 0:
            for l, i in zip(leads, range(ls)):
                X_train[:, :, i] = X1[:, :, l]  
                X_test[:, :, i] = X2[:, :, l]
                X_val[:, :, i] = X3[:, :, l]  
        else:  
            X_train = X1
            X_test = X2
            X_val = X3        
    #########################################################
        num_classes = 2
        input_shape = (num_samp, num_leads) # <<< ---------------------------
        batch_size = 64
        epochs = 80

        input_layer = Input(shape=(num_samp, num_leads))

        # conv1 = Conv1D(filters=32, kernel_size=5, padding='same')(input_layer)
        conv1 = Conv1D(filters=32, kernel_size=5, padding='same', use_bias=False)(input_layer)
        leakyrelu1 = LeakyReLU()(conv1)
        batch_norm1 = BatchNormalization()(leakyrelu1)
        pooling1 = MaxPooling1D(pool_size=2)(batch_norm1)

        # conv2 = Conv1D(filters=64, kernel_size=5, padding='same')(pooling1)
        conv2 = Conv1D(filters=64, kernel_size=5, padding='same', use_bias=False)(pooling1)
        leakyrelu2 = LeakyReLU()(conv2)
        batch_norm2 = BatchNormalization()(leakyrelu2)
        pooling2 = MaxPooling1D(pool_size=2)(batch_norm2)

        # conv3 = Conv1D(filters=64, kernel_size=5, padding='same')(pooling2)
        conv3 = Conv1D(filters=64, kernel_size=5, padding='same', use_bias=False)(pooling2)
        leakyrelu3 = LeakyReLU()(conv3)
        batch_norm3 = BatchNormalization()(leakyrelu3)
        pooling3 = MaxPooling1D(pool_size=2)(batch_norm3)

        # conv4 = Conv1D(filters=128, kernel_size=5, padding='same')(pooling3)
        conv4 = Conv1D(filters=128, kernel_size=5, padding='same', use_bias=False)(pooling3)
        leakyrelu4 = LeakyReLU()(conv4)
        batch_norm4 = BatchNormalization()(leakyrelu4)
        pooling4 = MaxPooling1D(pool_size=2)(batch_norm4)

        # conv5 = Conv1D(filters=128, kernel_size=5, padding='same')(pooling4)
        conv5 = Conv1D(filters=128, kernel_size=5, padding='same', use_bias=False)(pooling4)
        leakyrelu5 = LeakyReLU()(conv5)
        batch_norm5 = BatchNormalization()(leakyrelu5)
        pooling5 = MaxPooling1D(pool_size=2)(batch_norm5)

        dropout5 = Dropout(0.2)(pooling5)

        flatten = Flatten()(dropout5)
      
    ###############################################################    
    # Za pokretanje model bez additional features inputa
    
        dense = Dense(16, activation='relu')(flatten)
        output_layer = Dense(1, activation='sigmoid')(dense)
        model = Model(inputs=input_layer, outputs=output_layer)
    
    ###############################################################
        # additional_input = Input(shape=(3,), name='additional_features')

        # features_dense = Dense(16, activation='relu')(additional_input)
        # concatenated = concatenate([flatten, features_dense])

        # output_layer = Dense(1, activation='sigmoid')(concatenated)

        # model = Model(inputs=[input_layer, additional_input], outputs=output_layer)
    # ###############################################################
    
        model.compile(optimizer=Adam(), loss='binary_crossentropy', metrics=['accuracy'])
        model.summary()

##############################################################################################################################
#                                              MLFLOW SETUP
#############################################################################################################################
    
        mlflow.set_tracking_uri(uri="http://127.0.0.1:8081")
        mlflow.set_experiment("PTBXL_binarne_klasifikacije - RERUNS")
        mlflow.tensorflow.autolog(log_models=False, checkpoint=False)

        input_schema = Schema([
            TensorSpec(np.dtype(np.float64), (-1, num_samp, num_leads), "input_layer")
            # ,
            # TensorSpec(np.dtype(np.float32), (-1, 3), "additional_input")
        ])

        output_schema = Schema([
            TensorSpec(np.dtype(np.float32), (-1, 1), "output")
        ])

        signature = ModelSignature(inputs=input_schema, outputs=output_schema)

        input_example = {
            "input_layer": X_train[:3, :]
            # ,
            # "additional_input": features_train[:3]
        }
        run_name = "HYP_Binarni_klasifikator_100hz_bez_biasa_bez_atributa"

        ##################### TRENIRANJE MODELA                 ################################################################################    
        with mlflow.start_run(run_name=run_name, log_system_metrics=True) as run:
            lr = 0.001  
            reduce_lr = ReduceLROnPlateau(monitor='val_loss', factor=0.1, patience=10, min_lr=0.00001, verbose=1)
            history = model.fit(
                X_train, 
                # [X_train, features_train],
                y_train, 
                epochs=epochs, 
                batch_size=batch_size,
                validation_data=(X_val, y_val),
                # validation_data=([X_val, features_val], y_val),
                callbacks=[reduce_lr] 
            )

            mlflow.tensorflow.log_model(
                model,
                "tensorflow-example",
                input_example=input_example,
                signature=signature,
                pip_requirements=["-r requirements.txt"],
            )    
            now = datetime.now()
            timestamp = now.strftime("%Y%m%d_%H%M%S")   

            plt.plot(history.history['accuracy'], label='accuracy')
            plt.plot(history.history['val_accuracy'], label = 'val_accuracy')
            plt.xlabel('Epoch')
            plt.ylabel('Accuracy')
            plt.ylim([0.5, 1])
            plt.legend(loc='lower right')
            history_plot_path = f"{output_directory}/history_plot_{timestamp}.png"
            plt.savefig(history_plot_path)
            mlflow.log_artifact(history_plot_path)

            model_name = f"OTHERS_HYP_binary_cnn_model_{timestamp}.h5"
            model.save(f"{output_directory}/{model_name}")
            mlflow.log_artifact(f"{output_directory}/{model_name}")

            ##################### UCITAVANJE ISTRENIRANOG MODELA    ###########################################################################################
            # model = load_model(
            #     model_name)
            ###############################################################################################################################    

            ##################### TESTIRANJE MODELA                 ################################################################################    

            test_loss, test_acc = model.evaluate(X_test,  y_test, verbose=2)
            # test_loss, test_acc = model.evaluate([X_test, features_test],  y_test, verbose=2)
            mlflow.log_metric("test_loss", test_loss)
            mlflow.log_metric("test_accuracy", test_acc)
            print(f'Accuracy: {test_acc * 100:.2f}%')


            predicted_probabilities = model.predict(X_test)
            # predicted_probabilities = model.predict([X_test, features_test])
            predicted_classes = (predicted_probabilities >= 0.5).astype(int)
            predicted_labels = [class1 if pred == 1 else class0 for pred in predicted_classes]
            y_pred = predicted_labels

            mapping_dict = {
            class0: 0,
            class1: 1
            }
            outputs = np.array([mapping_dict[val] for val in predicted_labels])

            print(f'Accuracy: {test_acc * 100:.2f}%')

            f1 = f1_score(y_test, outputs, average='binary')
            mlflow.log_metric("f1_score", f1)
            print(f'F1 Score: {f1* 100:.2f}%')

            auroc = roc_auc_score(y_test, predicted_probabilities)
            mlflow.log_metric("auroc_score", auroc)
            print(f'AUROC Score: {auroc:.4f}')

            print("Classification Report:")
            classification_rep = classification_report(y_test, outputs, zero_division=0)
            mlflow.log_text(classification_rep, "classification_report.txt")
            print(classification_rep)

            output_file = f'{output_directory}/metrics_report{timestamp}.txt'
            with open(output_file, "w") as file:
                file.write(f"Accuracy: {test_acc * 100:.2f}%\n")
                file.write(f"F1 Score: {f1 * 100:.2f}%\n")
                file.write(f"AUROC Score: {auroc:.4f}\n")
                file.write("Classification Report:\n")
                file.write(classification_rep)
                file.write("Classes: \n")
                file.write(f"Class 0 - {class0} Class 1 - {class1}")

            mlflow.log_artifact(output_file)
            print(f"Metrics have been written to {output_file}")

            class_names = [class0, class1]
            cm = confusion_matrix(y_test, outputs)
            plt.figure(figsize=(8, 6))
            sns.heatmap(cm, annot=True, fmt='d', cmap='Blues', cbar=False)
            plt.xticks(ticks=[0.5, 1.5], labels=class_names)
            plt.yticks(ticks=[0.5, 1.5], labels=class_names)
            plt.xlabel('Predicted Label')
            plt.ylabel('True Label')
            plt.title('Confusion Matrix')
            confusion_matrix_path = f'{output_directory}/confusion_matrix_{timestamp}.png'
            plt.savefig(confusion_matrix_path) 
            mlflow.log_artifact(confusion_matrix_path)



            fpr, tpr, thresholds = roc_curve(y_test,  predicted_probabilities)
            plt.figure(figsize=(8, 6))
            plt.plot(fpr, tpr, color='blue', lw=2, label='ROC curve (AUROC = %0.2f)' % auroc)
            plt.plot([0, 1], [0, 1], color='gray', linestyle='--')
            plt.xlabel('False Positive Rate')
            plt.ylabel('True Positive Rate')
            plt.title('Receiver Operating Characteristic (ROC) Curve')
            plt.legend(loc='lower right')
            plt.grid(True)
            roc_curve_path = f'{output_directory}/ROC_curve_{timestamp}.png'
            plt.savefig(roc_curve_path) 
            mlflow.log_artifact(roc_curve_path)

