import random
import pandas as pd
import numpy as np
import seaborn as sns
import matplotlib.pyplot as plt
import tensorflow as tf
import joblib, sys, os
from sklearn.metrics import f1_score, confusion_matrix, roc_auc_score, roc_curve, auc,  multilabel_confusion_matrix, classification_report
from tensorflow.keras.models import Sequential, Model, load_model
from tensorflow.keras.layers import Conv1D, MaxPooling1D, Flatten, Dense, Dropout, Input, LeakyReLU, Softmax, concatenate, BatchNormalization
from tensorflow.keras.utils import to_categorical
from sklearn.preprocessing import MultiLabelBinarizer
from keras.callbacks import ReduceLROnPlateau
from datetime import datetime
from tensorflow.python.client import device_lib
import mlflow
from mlflow.models import ModelSignature
from mlflow.types import Schema, TensorSpec

print("device LIST    ", device_lib.list_local_devices())
gpus = tf.config.experimental.list_physical_devices('GPU')
print("Num GPUs Available: ", len(gpus))
for gpu in gpus:
  tf.config.experimental.set_memory_growth(gpu, True)
print("TensorFlow version:", tf.__version__)

def load_data(directory, file_name):
    path = os.path.join(directory, file_name)
    return joblib.load(path)    
    
def convert_to_string(lst):
    if not lst:
        return ''
    return ', '.join(lst)

def switch_places(entry):
    if ',' in entry:
        parts = entry.split(',')
        parts = [word.strip() for word in parts]
        parts.sort()  
        return ', '.join(parts)
    return entry

def calculate_label_weights(y):
    # koristi inevrznu frekvenciju
    #  vraca dictionary objekat tezina labela, gdje su kljucevi indeksi labela
    n_samples, n_labels = y.shape
    label_weights = {}
    
    for i in range(n_labels):
        pos_weight = (n_samples - np.sum(y[:, i])) / np.sum(y[:, i])
        label_weights[i] = pos_weight
    
    return label_weights

def weighted_binary_crossentropy(y_true, y_pred):
    # loss funkcija za weighted binary crossentorpiju za okruzenje u kom jedan primjerak ima vise labela
    loss = 0.0
    
    for i, weight in label_weights.items():
        y_true_label = y_true[:, i]
        y_pred_label = y_pred[:, i]

        label_loss = tf.reduce_mean(tf.nn.weighted_cross_entropy_with_logits(logits=y_pred_label, labels=y_true_label, pos_weight=weight))
        loss += label_loss
 
    loss /= len(label_weights)
    return loss

def plot_multilabel_roc(y_true, y_scores, num_classes, class_names, timestamp):
    #  posto je specifican slucaj multi class multi label klasifikacije treba posebna funkcija za vizuelizaciju ROC krive
    plt.figure(figsize=(10, 8))

    for i in range(num_classes):
        fpr, tpr, thresholds = roc_curve(y_true[:, i], y_scores[:, i])
        roc_auc = auc(fpr, tpr)

        plt.plot(fpr, tpr, linestyle='-', label='Class %s (area = %0.2f)' % (class_names[i], roc_auc))

    plt.plot([0, 1], [0, 1], 'k--', label='Random (area = 0.5)')

    plt.xlabel('False Positive Rate')
    plt.ylabel('True Positive Rate')
    plt.title('ROC Curve for Multi-label Classification')
    plt.legend(loc="lower right")
    roc_curve_path = f"{output_directory}/ROC_for_multilabel_super_{timestamp}.png"
    plt.savefig(roc_curve_path) 
    mlflow.log_artifact(roc_curve_path)

if __name__ == "__main__":
    input_directory = "../records_ptbxl"  
    output_directory = "multiclass_subclass_classification_model_output"
    if not os.path.exists(output_directory):
        os.makedirs(output_directory)
        
    leads = [] # odje navodimo kanale (indekse od 0 do 11) koje zemilo da koristimo; ako je lista prazna podrazumijeva se da svaki kanala ucestvuje

    signals1 = load_data(input_directory, "X_train.sav")
    X1 = np.array([item['signal'] for item in signals1])
    labels1 = load_data(input_directory, "y_train.sav")
    y_train = [item['label_sub'] for item in labels1]
    features1 = load_data(input_directory, "features_train.sav")
    features_train = np.array([item['features'] for item in features1])

    signals2 = load_data(input_directory, "X_test.sav")
    X2 = np.array([item['signal'] for item in signals2])
    labels2 = load_data(input_directory, "y_test.sav")
    y_test = [item['label_sub'] for item in labels2]
    features2 = load_data(input_directory, "features_test.sav")
    features_test = np.array([item['features'] for item in features2])

    signals3 = load_data(input_directory, "X_val.sav")
    X3 = np.array([item['signal'] for item in signals3])
    labels3 = load_data(input_directory, "y_val.sav")
    y_val = [item['label_sub'] for item in labels3]
    features3 = load_data(input_directory, "features_val.sav")
    features_val = np.array([item['features'] for item in features3])
    print(len(y_train))
    print(y_train)


    mlb = MultiLabelBinarizer()
    y_train_binary = mlb.fit_transform(y_train)
    y_train_binary = y_train_binary.astype('float32')

    y_test_binary = mlb.transform(y_test)
    y_test_binary = y_test_binary.astype('float32')

    y_val_binary = mlb.transform(y_val)
    y_val_binary = y_val_binary.astype('float32')

    features_train = np.asarray(features_train).astype('float32')
    features_test = np.asarray(features_test).astype('float32')
    features_val = np.asarray(features_val).astype('float32')

    num_samp = 1000
    ls = len(leads)
    if(ls == 0):
        num_leads = 12
    else:
        num_leads = ls

    X_train = np.zeros((X1.shape[0], num_samp, num_leads))
    X_test = np.zeros((X2.shape[0], num_samp, num_leads))
    X_val = np.zeros((X3.shape[0], num_samp, num_leads))

############################################################################
# ovdje se selektuju kanali koji ce ucestvovati u treniranju i testiranju
    if ls != 0:
        for l, i in zip(leads, range(ls)):
            X_train[:, :, i] = X1[:, :, l]  
            X_test[:, :, i] = X2[:, :, l]
            X_val[:, :, i] = X3[:, :, l]  
    else:  
        X_train = X1
        X_test = X2
        X_val = X3        
#########################################################
    num_classes = 21
    input_shape = (num_samp, num_leads) 
    batch_size = 64
    epochs = 80

    input_layer = Input(shape=(num_samp, num_leads))

    conv1 = Conv1D(filters=32, kernel_size=5, padding='same', use_bias=False)(input_layer)
    leakyrelu1 = LeakyReLU()(conv1)
    batch_norm1 = BatchNormalization()(leakyrelu1)
    # dropout1 = Dropout(0.2)(batch_norm1)  
    pooling1 = MaxPooling1D(pool_size= 2)(batch_norm1)

    conv2 = Conv1D(filters=64, kernel_size=5, padding='same', use_bias=False)(pooling1)
    leakyrelu2 = LeakyReLU()(conv2)
    batch_norm2 = BatchNormalization()(leakyrelu2)
    # dropout2 = Dropout(0.2)(batch_norm2)
    pooling2 = MaxPooling1D(pool_size= 2)(batch_norm2)

    conv3 = Conv1D(filters=64, kernel_size=5, padding='same', use_bias=False)(pooling2)
    leakyrelu3 = LeakyReLU()(conv3)
    batch_norm3 = BatchNormalization()(leakyrelu3)
    # dropout3 = Dropout(0.2)(batch_norm3) 
    pooling3 = MaxPooling1D(pool_size= 2)(batch_norm3)

    conv4 = Conv1D(filters=128, kernel_size=5, padding='same', use_bias=False)(pooling3)
    leakyrelu4 = LeakyReLU()(conv4)
    batch_norm4 = BatchNormalization()(leakyrelu4)
    # dropout4 = Dropout(0.2)(batch_norm4)  
    pooling4 = MaxPooling1D(pool_size= 2)(batch_norm4)

    conv5 = Conv1D(filters=128, kernel_size=5, padding='same', use_bias=False)(pooling4)
    leakyrelu5 = LeakyReLU()(conv5)
    batch_norm5 = BatchNormalization()(leakyrelu5)
    pooling5 = MaxPooling1D(pool_size= 2)(batch_norm5)

    dropout5 = Dropout(0.2)(pooling5) 
    flatten = Flatten()(dropout5)

###############################################################    
    # Za pokretanje modela bez additional features inputa    

    # dense = Dense(units=num_classes)(flatten) 
    # output_layer = Softmax()(dense)
    # model = Model(inputs=input_layer, outputs=output_layer)

###############################################################
    processed_additional_input = Input(shape=(3,), name='additional_features')
    features_dense = Dense(8, activation='relu')(processed_additional_input)
    concatenated = concatenate([flatten, features_dense])
    dense = Dense(units=num_classes)(concatenated) 
    
    output_layer = Softmax()(dense)
    model = Model(inputs=[input_layer, processed_additional_input], outputs=output_layer)

##################################################################################33
    model.compile(optimizer='adam', loss=weighted_binary_crossentropy, metrics=['accuracy'])
    model.summary()

##############################################################################################################################
#                                              MLFLOW SETUP
#############################################################################################################################
        
    mlflow.set_tracking_uri(uri="http://127.0.0.1:8081")
    mlflow.set_experiment("PTBXL_multiklasifikacije")
    mlflow.tensorflow.autolog(log_models=False, checkpoint=False)

    input_schema = Schema([
        TensorSpec(np.dtype(np.float64), (-1, num_samp, num_leads), "input_layer")
        ,
        TensorSpec(np.dtype(np.float32), (-1, 3), "additional_input")
    ])

    output_schema = Schema([
        TensorSpec(np.dtype(np.float32), (-1, 5), "output")
    ])

    signature = ModelSignature(inputs=input_schema, outputs=output_schema)

    input_example = {
        "input_layer": X_train[:3, :]
        ,
        "additional_input": features_train[:3]
    }
    run_name= "Subclass_Multiklasifikator_100hz_sa_atributima_bez_biasa_21_klasa"

##################### TRENIRANJE MODELA                 ################################################################################    
    with mlflow.start_run(run_name=run_name, log_system_metrics=True) as run:
              
        lr = 0.001
        label_weights = calculate_label_weights(y_train_binary)
        reduce_lr = ReduceLROnPlateau(monitor='val_loss', factor=0.1, patience=10, min_lr=0.00001, verbose=1)

        # history = model.fit(
        #     X_train, 
        #     y_train_binary, 
        #     epochs=epochs, 
        #     batch_size=batch_size, 
        #     validation_data=(X_val, y_val_binary),
        #     callbacks=[reduce_lr],
        #     verbose=1
        # )   

        history = model.fit(
            [X_train, features_train], 
            y_train_binary, 
            epochs=epochs, 
            batch_size=batch_size, 
            validation_data=([X_val, features_val], y_val_binary),
            callbacks=[reduce_lr],
            verbose=1
        )         

        mlflow.tensorflow.log_model(
                model,
                "tensorflow-example",
                input_example=input_example,
                signature=signature,
                pip_requirements=["-r requirements.txt"],
        )   

        now = datetime.now()
        timestamp = now.strftime("%Y%m%d_%H%M%S")   

        plt.plot(history.history['accuracy'], label='accuracy')
        plt.plot(history.history['val_accuracy'], label = 'val_accuracy')
        plt.xlabel('Epoch')
        plt.ylabel('Accuracy')
        plt.ylim([0.5, 1])
        plt.legend(loc='lower right')
        history_plot_path = f"{output_directory}/history_plot_super_cl_{timestamp}.png"
        plt.savefig(history_plot_path)
        mlflow.log_artifact(history_plot_path)

        model_name = f"multiclass_classification_cnn_super_cl_{timestamp}.h5"
        model_path = f"{output_directory}/{model_name}" 
        model.save(model_path)
        mlflow.log_artifact(model_path)

        ##################### UCITAVANJE ISTRENIRANOG MODELA    ###########################################################################################
        # label_weights = calculate_label_weights(y_train_binary)
        # model = load_model(
        #     model_name, 
        #     custom_objects={
        #     'weighted_binary_crossentropy': weighted_binary_crossentropy}
        # )
        ########################################################################################################################################    
        ##################### TESTIRANJE MODELA             #################################################################################
        
        # test_loss, test_acc = model.evaluate(X_test,  y_test_binary, verbose=2)
        test_loss, test_acc = model.evaluate([X_test, features_test],  y_test_binary, verbose=2)
        mlflow.log_metric("test_loss", test_loss)
        mlflow.log_metric("test_accuracy", test_acc)
        print(f'Accuracy: {test_acc * 100:.2f}%')

        class_names = mlb.classes_
        safe_class_names = [name.replace('/', '_') for name in class_names]

        # predicted_probabilities = model.predict(X_test)
        predicted_probabilities = model.predict([X_test, features_test])
        outputs = (predicted_probabilities >= 0.5).astype(int)

        f1 = f1_score(y_test_binary, outputs, average='weighted')
        mlflow.log_metric("f1_score", f1)
        print(f'F1 Score: {f1* 100:.2f}%')

        auroc = roc_auc_score(y_test_binary, predicted_probabilities)
        mlflow.log_metric("auroc_score", auroc)
        print(f'AUROC Score: {auroc:.4f}')

        print("Classification Report:")
        classification_rep = classification_report(y_test_binary, outputs, zero_division=0)
        mlflow.log_text(classification_rep, "classification_report.txt")
        print(classification_rep)

        output_file = f'{output_directory}/metrics_report{timestamp}.txt'
        with open(output_file, "w") as file:
            file.write(f"Accuracy: {test_acc * 100:.2f}%\n")
            file.write(f"F1 Score: {f1 * 100:.2f}%\n")
            file.write(f"AUROC Score: {auroc:.4f}\n")
            file.write("Classification Report:\n")
            file.write(classification_rep)
            file.write("Classes: \n")
            file.write(f" Superclasses: {class_names}")
        mlflow.log_artifact(output_file)
        print(f"Metrics have been written to {output_file}")

        y_pred = mlb.inverse_transform(outputs)
        mcm = multilabel_confusion_matrix(y_test_binary, outputs)


        for i in range(num_classes):
            fig, ax = plt.subplots(figsize=(8, 6))  
            cm = mcm[i]
            sns.heatmap(cm, annot=True, fmt='d', cmap='Purples', cbar=False, ax=ax)  
            ax.set_title(f'Confusion Matrix for Class {class_names[i]}')
            ax.set_xlabel('Predicted Label')
            ax.set_ylabel('True Label')

            plt.tight_layout() 
            confusion_matrix_path = f"{output_directory}/confusion_matrix_{safe_class_names[i]}_super_cl_{timestamp}.png"
            plt.savefig(confusion_matrix_path) 
            mlflow.log_artifact(confusion_matrix_path)
            plt.close(fig) 

        plot_multilabel_roc(y_test_binary, predicted_probabilities, num_classes, class_names, timestamp)    

